# -*- coding: utf-8 -*-
__all__ = [
    "Catch22ForestClassifier",
    "HIVECOTEV1",
]

from sktime.classification.hybrid._catch22_forest_classifier import (
    Catch22ForestClassifier,
)
from sktime.classification.hybrid._hivecote_v1 import HIVECOTEV1
