#!/usr/bin/env python3 -u
# -*- coding: utf-8 -*-
# copyright: sktime developers, BSD-3-Clause License (see LICENSE file)

"""Base class for defining performance metrics."""

__author__ = ["Ryan Kuhns"]
__all__ = ["BaseMetric"]

from sktime.performance_metrics.base._base import BaseMetric
