print("import: 'sktime'")
import sktime

print("import: 'sktime.classification'")
import sktime.classification

print("import: 'sktime.datasets'")
import sktime.datasets

print("import: 'sktime.distances'")
import sktime.distances

print("import: 'sktime.forecasting'")
import sktime.forecasting

print("import: 'sktime.performance_metrics'")
import sktime.performance_metrics

print("import: 'sktime.regression'")
import sktime.regression

print("import: 'sktime.series_as_features'")
import sktime.series_as_features

print("import: 'sktime.transformations'")
import sktime.transformations

print("import: 'sktime.utils'")
import sktime.utils

